import json
import logging

from .arguments import parse_arguments
from .exception import DudendasArgumentException
from .logging import setup_logger
from .search import DudenSearch

logger = logging.getLogger("dudendas")


def get_words(keyword):
    msg = "Searching keyword '%s'."
    logger.debug(msg, keyword)

    search = DudenSearch(keyword)
    search.parse_results()
    words = search.parse_words(group=keyword)

    msg = "Got %d words with keyword '%s'."
    logger.info(msg, len(words), keyword)

    return words


def main():
    setup_logger(logger)

    try:
        args = parse_arguments()
    except DudendasArgumentException as e:
        msg = 'Failed to parse arguments: ' + str(e)
        logger.error(msg)
        exit(1)

    if args.debug:
        logger.setLevel(logging.DEBUG)
        msg = 'Application is running in debug mode.'
        logger.debug(msg)
    elif args.quiet:
        logger.setLevel(logging.WARNING)

    logger.info('Starting Dudendas.')

    dumpfile = args.file
    keywords = args.keywords

    try:
        words = set()

        for keyword in keywords:
            new_words = get_words(keyword)
            words.update(new_words)

        with open(dumpfile, "w") as f:
            json.dump([w.serialize() for w in words], f, indent="\t")

    except KeyboardInterrupt:
        msg = 'Received keyboard interrupt.'
        logger.info(msg)
        exit(1)



import logging

from argparse import ArgumentParser

from .exception import DudendasArgumentException

logger = logging.getLogger("dudendas")


def parse_arguments():
    parser = ArgumentParser(
        prog="Dudendas",
        description="A tool for collecting words from Duden."
    )
    parser.add_argument("-v", "--debug", action="store_true",
                        help="Print debug information.")
    parser.add_argument("-q", "--quiet", action="store_true",
                        help="Print errors and warnings only.")
    parser.add_argument("-f", "--file", type=str, required=True,
                        help="File to dump fetched content into.")
    parser.add_argument("keywords", type=str, nargs='+',
                        help="Keywords to search for.")

    args = parser.parse_args()

    if args.debug and args.quiet:
        msg = 'Cannot be quiet in debug mode.'
        raise DudendasArgumentException(msg)

    return args



import logging
import re

logger = logging.getLogger("dudendas")


def decompose_all(obj, name):
    for item in obj.find_all(name):
        item.decompose()


def parse_list(root):
    if root.name in ["ul", "ol"]:
        list_ = root
    else:
        list_ = root.ol or root.ul

    if list_ is not None:
        items = list_.find_all("li", recursive=False)
        return [parse_list(i) for i in items]
    else:
        text = root.get_text()
        return textify(text)


def textify(s):
    s = re.sub(r"\([a-z0-9][a-z]?\)", "", s)
    s = re.sub(r"\s{2,}", " ", s)
    s = s.strip()

    return s


def flatten_list(root):
    items = list()

    for item in root:
        if type(item) is list:
            items.extend(flatten_list(item))
        else:
            items.append(item)

    return items



import aiohttp
import asyncio
import logging

from bs4 import BeautifulSoup
from urllib.request import urlopen

from .util import decompose_all

logger = logging.getLogger("dudendas")


def build_soup(page):
    page = page.replace("\u00ad", "")
    soup = BeautifulSoup(page, "html.parser")
    main = soup.find("main", id="content")
    decompose_all(main, "script")

    return main


def load_page(url):
    with urlopen(url) as page:
        page = page.read()

    soup = build_soup(page.decode("UTF-8"))

    return soup


async def _fetch_url(session, url, retries=5):
    content = None

    for i in range(retries):
        try:
            async with session.get(url) as response:
                response.raise_for_status()
                content = await response.read()

        except aiohttp.ClientResponseError:
            msg = "Could not fetch '%s' [%d/%d]."
            logger.debug(msg, url, i + 1, retries)
            continue

        encoding = response.charset if response.charset else "ISO-8859-1"
        content = content.decode(encoding)
        break

    if content is None:
        msg = "Could not retrieve '{}'."
        logger.warning(msg.format(url))

    return url, content


async def _load_pages(urls, limit):
    def init_task(session, url):
        task = _fetch_url(session, url)
        return task

    connector = aiohttp.TCPConnector(limit=limit)

    async with aiohttp.ClientSession(connector=connector) as session:
        tasks = [init_task(session, url) for url in urls]
        results = await asyncio.gather(*tasks)

    dl_successful = results.count(True)
    dl_rate = dl_successful / float(len(results))
    msg = "total: {}, successful: {}, rate: {}"
    logger.info(msg.format(len(results), dl_successful, dl_rate))

    return results


def load_pages(urls, limit=100):
    msg = "Downloading {} files."
    logger.info(msg.format(len(urls)))

    loop = asyncio.get_event_loop()
    func = _load_pages(urls, limit=limit)
    contents = loop.run_until_complete(func)

    return contents



import logging
import re

from urllib.parse import urlparse, parse_qs

from .network import build_soup, load_page, load_pages
from .word import DudenWord

logger = logging.getLogger("dudendas")


class DudenSearch(object):
    HREF_PATTERN = re.compile(r"https://www\.duden\.de/rechtschreibung/.*")
    query = "https://www.duden.de/suchen/dudenonline/{keyword}?page={page}"

    def __init__(self, keyword):
        self.keyword = keyword
        self.words = set()
        self.results = set()
        self.page_count = self._get_page_count()

    def __repr__(self):
        fmt = "<DudenSearch '{}'>"
        return fmt.format(self.keyword)

    def parse_results(self):
        urls = self._build_urls(self.page_count)

        msg = 'Start parsing {} results.'
        logger.debug(msg.format(len(urls)))

        for k, result in enumerate(load_pages(urls)):
            url, content = result

            soup = build_soup(content)

            try:
                logger.info("Loading page [%d/%d].", k + 1, self.page_count + 1)
                links = soup.find_all("a", href=self.HREF_PATTERN)

                msg = 'Found {} matches on page {}.'
                logger.debug(msg.format(len(links), k + 1))

                for l in links:
                    href = l["href"]
                    self.results.add(href)
            except Exception as e:
                logger.warning("Caught %s for page %d.", type(e).__name__, k + 1)

        return self.results

    def parse_words(self, group=None):
        msg = 'Start parsing {} words.'
        logger.debug(msg.format(len(self.results)))

        for k, result in enumerate(load_pages(self.results)):
            url, content = result

            soup = build_soup(content)
            name = DudenSearch._name_from_href(url)

            logger.info("Parsing %s [%d/%d].", name, k + 1, len(self.results))

            try:
                word = DudenWord(name, url, soup)
                if word.belongs_to(group):
                    self.words.add(word)
            except Exception as e:
                logger.warning("Caught %s for word '%s'.", type(e).__name__, name)

        return self.words

    def _build_urls(self, page_count):
        r = range(0, self.page_count + 1)
        urls = [self.query.format(keyword=self.keyword, page=p) for p in r]
        return urls

    def _get_page_count(self):
        soup = self._load_page(0)

        pager_last = soup.find("span", "pager-last").a

        if pager_last is None:
            pages = 1
        else:
            tmp_url = pager_last["href"]
            query_raw = urlparse(tmp_url).query
            query = parse_qs(query_raw)
            pages = int(query["page"][0])

        msg = 'Page count is {}.'
        logger.debug(msg.format(pages))

        return pages

    def _load_page(self, page):
        url = self.query.format(keyword=self.keyword, page=page)

        msg = "Loading page '{}'."
        logger.debug(msg.format(url))

        return load_page(url)

    @staticmethod
    def _name_from_href(href):
        path = urlparse(href).path
        word = path.split("/")[-1]
        return word



import logging
import re

from .exception import DudendasException
from .util import decompose_all, parse_list, flatten_list

logger = logging.getLogger("dudendas")


class DudenWord(object):
    def __init__(self, word, url, soup):
        self.word = word
        self.url = url

        self.title = self._get_title(soup)
        self.meaning = self._get_meaning(soup)
        self._block = soup.find(id="block-system-main")

        if self._block is None:
            msg = "Cannot find the main section for '{}'."
            raise DudendasException(msg.format(self.word))

    def __repr__(self):
        fmt = "<DudenWord '{}'>"
        return fmt.format(self.word)

    def __eq__(self, other):
        return self.word == other.word

    def __hash__(self):
        return hash(self.word)

    def belongs_to(self, group):
        for s in self._block.find_all("strong"):
            if s.string is not None and group in s.string:
                return True

        for m in flatten_list(self.meaning):
            if re.search(r"\(.*" + group + r".*\)", m):
                return True
            elif re.search(group + " für", m):
                return True

        logger.debug("Word '%s' does not belong to group '%s'.", self.word, group)

        return False

    def serialize(self):
        meaning = flatten_list(self.meaning)
        meaning = "\n".join(meaning)

        return dict(word=self.word,
                    title=self.title,
                    meaning=meaning,
                    url=self.url)

    def _get_title(self, soup):
        h1 = soup.h1.string
        return h1.strip()

    def _get_meaning(self, soup):
        section = self._find_section(soup, "Bedeutungsübersicht")

        if section is None:
            return ""

        for name in ["header", "section", "figure"]:
            decompose_all(section, name)

        return [parse_list(section)]

    def _find_section(self, soup, title, approximate=False):
        for section in soup.find_all("section"):
            if section.h2:
                if section.h2.text == title:
                    return section
                elif approximate and title in section.h2.text:
                    return section

        return None
